#ifndef __LINBOX_INTEGER_SMITH_NORMAL_FORM_C__
#define __LINBOX_INTEGER_SMITH_NORMAL_FORM_C__
// ======================================================================= //
// Copyright (C) LinBox project 1999
// Givaro / Athapascan-1
// Smith form
// Time-stamp: <08 Dec 11 12:15:49 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //

#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include <time.h>
#include <list>

#include "givintfactor.h"
#include "givintrns.h"
#include "givgfq.h"


#include "LinBox/lin_dom_gauss.C"
#include "LinBox/lin_pp_gauss.C"
#include "LinBox/lin_density.h"
#include "LinBox/lin_invariants.h"
#include "LinBox/lin_spv_bb.h"
#include "LinBox/lin_erase_bb.h"
#include "LinBox/lin_wrap_c++.h"


typedef SparseBlackBoxDom< OperatorWrapper< GFqDomain::Residu_t > > SPBBlong;
typedef SparseBlackBoxEraseDom< GFqDomain > SPBBE ;
    
void one_gauss(unsigned long& rank, char * Matrix_File_Name, GFqDomain::Residu_t MOD) {
//std::cerr << "MFN: " << Matrix_File_Name << ", MOD: " << MOD << std::endl;
    
    GFqDomain F(MOD);

    SPBBE TheMat(F);
    unsigned long ni(0), nj(0), addrank; TheMat.read(ni,nj,Matrix_File_Name);
    if (ni > nj) {
        addrank = TheMat.read_erase(Matrix_File_Name);
        char *UT, * File_Name;
        File_Name = tempnam("/tmp","bbx_"); // "/tmp/bbx_XXXXXX"; mkstemp(File_Name);
        TheMat.write(File_Name);
        TheMat.read_transpose(File_Name);
        UT = new char[5+strlen(File_Name)];
        sprintf(UT,"\\rm %s", File_Name);
        system(UT);
        delete [] UT; delete [] File_Name;
    } else
        addrank = TheMat.read_erase(Matrix_File_Name);

    char *UT = new char[40];
    sprintf(UT,"additional rank : %ld", addrank);
    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
    delete [] UT; 
   
    
    GaussDom< GFqDomain > GF( F, GlobalCommentator );    
    GF.gauss_rankin(rank,TheMat,Density());
    if (addrank > 0) rank+=addrank;
    
//std::cerr << "addrank: " << addrank << ", rank: " << rank << std::endl;
}

template<class InvClass>
void one_permuts(InvClass& c_rows, unsigned long& rank, char * Matrix_File_Name, GFqDomain::Residu_t MOD) {
    
    GFqDomain F(MOD);
    GaussDom< GFqDomain > GF( F, GlobalCommentator );    

    SPBBE TheMat(F);

    TheMat.read_transpose(Matrix_File_Name);

    GF.gauss_rankin(rank,TheMat,Density(), 1);

    for(unsigned long i = 0; i< TheMat.size(); ++i)
        if (TheMat[i].size() > 0) c_rows.PushInvariant(i, 1);
}

#define GIVARO_MACHINT GFqDomain::element
#define GIVARO_MACHIND unsigned long
void one_pp_gauss(unsigned long& rank, char * Matrix_File_Name, GFqDomain::element PRIMEPOWER, GFqDomain::element PRIME) {
        OperatorWrapper<GIVARO_MACHINT> F;
        SparseBlackBoxDom< OperatorWrapper<GIVARO_MACHINT>, GIVARO_MACHIND >  MF( F );
        SparseBlackBoxDom< OperatorWrapper<GIVARO_MACHINT>, GIVARO_MACHIND >::element TheMat;
        GIVARO_MACHIND ni(0), nj(0), nz(0);
        MF.read(ni, nj, Matrix_File_Name);
        if (ni > nj)
            MF.read_transpose(TheMat, ni, nj, nz, Matrix_File_Name);
        else
            MF.read(TheMat, ni, nj, nz, Matrix_File_Name);
	
	// MF.write( "Toto", TheMat );

        prime_power_rankin(PRIMEPOWER, PRIME, rank, TheMat, ni, nj, Density());
}


// void gauss_Integer_rank(ulong& rank, Integer& p1, char * File_Name, Integer IN, double& wrong, double error) {
//     IntPrimeDom IPD;
//     IPD.prevprime(p1, 1+MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE/20);
//     while ( IPD.iszero(IN % p1)) IPD.prevprimein(p1);
//     one_gauss(rank, File_Name, (long)p1);
// }
void gauss_Integer_rank(ulong& rank, Integer& p1, char * File_Name, Integer IN, double& wrong, double error) {
    IntPrimeDom IPD;
    p1 = Integer(2);
    while ( IPD.iszero(IN % p1)) IPD.nextprimein(p1);
    one_gauss(rank, File_Name, (GFqDomain::Residu_t) p1);
}



template<class InvClass>
void gauss_rank_N_torsion(InvClass& Inv, char * File_Name, const Integer& IN, bool withintegerrank = 0) {
    startActivity(GlobalCommentator, "N-Torsion", "Rank N-Torsion Computation", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    Integer valence;
    ulong nbfactor;
    std::list<Integer> Lm;
    std::list<ulong> Le;
    
    IntFactorDom<> IFD;
    nbfactor = IFD.set(Lm,Le,IN);

    typename std::list<Integer>::iterator im=Lm.begin();
    typename std::list<ulong>::iterator ie=Le.begin();

    SPBBlong::Domain_t LongDom;
    SPBBlong TheMat( LongDom );
    unsigned long ni(0), nj(0); TheMat.read(ni,nj,File_Name);

    unsigned long * r = new unsigned long[nbfactor];
    bool * t = new bool[nbfactor];
    unsigned long ir;
    unsigned long k, dimr = GIVMIN(ni,nj);
    unsigned long minr = dimr, maxr = 0;
    
    char * UT = new char[170]; 


    Integer pr=1;
    for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
        if (*ie > 0) {
            pr *= *im;
            if (*im < Integer(MAXPRIMESIZE)) {
                t[k] = true;
                one_gauss( r[k], File_Name, (GFqDomain::Residu_t)(*im));
                minr = GIVMIN(minr,r[k]);
                maxr = GIVMAX(maxr,r[k]);
                
               std::string s; Integer2string(s, *im);
               sprintf(UT,"rank mod %s : %ld", s.c_str(), r[k]);
               activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
            } else {
                t[k] = false;
                std::string s; Integer2string(s, *im);
                sprintf(UT,"WARNING, %s is TOO BIG. Supposed full rank, WARNING.", s.c_str());
                activityReport(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_WARNING);
            }
           
        }
    
    if (withintegerrank) {
        Integer p(2);
        while ((iszero(pr % p)) || (! IFD.isprime(p)))
            IFD.nextprimein(p);
        one_gauss( maxr, File_Name, (GFqDomain::Residu_t)p);
        std::string s; Integer2string(s, p);
//         GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT ) << "rank mod " << (GFqDomain::Residu_t)(p) << " : " << maxr << endl;
        sprintf(UT,"Integer rank (mod %s) : %ld", s.c_str() , maxr);
        activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        ir = maxr;
    } else {
        ir = dimr;
    }
    
    
    if (ir > minr) {
        unsigned long ssize = ir-minr;
        GFqDomain::Residu_t * Smith = new GFqDomain::Residu_t[ssize];
        unsigned long ll=ssize;
        for (;ll--;)
            Smith[ll] = 1;
    

        unsigned long l=0, count_ie;
        for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
            if (t[k]) {
                if (*ie > 0) {
                    count_ie = 0;
                    l = r[k]; 
                    GFqDomain::Residu_t p = (GFqDomain::Residu_t)(*im);
                    GFqDomain::Residu_t pp = p;
//                 while (l < ir) {
                    while ( (l<ir) || (!withintegerrank) ) {
                        for(unsigned long lk=ir-l;lk;--lk)
                            Smith[ssize-lk] *= p;
                        pp *= p;
                        if (++count_ie < *ie) {
                            one_pp_gauss(l, File_Name, pp,p);
                            maxr = GIVMAX(maxr, l);
                        } else
                            break;
//                         GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT ) << "pp rank mod " << pp << " : " << l << endl;
                        std::string s; Integer2string(s,pp);
                        sprintf(UT,"pp rank mod %s : %ld", s.c_str(), l);
                        activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
                    }
                }
            }

        Inv.PushInvariant(1,minr);
        unsigned long lj=0;
        ssize = GIVMIN(maxr, dimr) - minr;
        for(ll=0;ll<ssize;) {
            while ((++lj<ssize) && (Smith[lj] == Smith[ll])) {};
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
//         delete [] Smith;
    
    } else {
        Inv.PushInvariant(1,ir);
    }
    
        
//     delete [] r; delete [] t; delete [] UT;
    stopActivity(GlobalCommentator, "Done.", LVL_BLABLA, INTERNAL_DESCRIPTION ); 
};


template<class InvClass>
void Smith2Invariant(unsigned long ir, unsigned long minr, InvClass& Inv, const std::vector< GFqDomain::Residu_t >& Smith) {
    if (ir > minr) {
        Inv.PushInvariant(1,minr);
        unsigned long lj=minr;
        for(long ll=minr;ll<Smith.size();) {
            while ((++lj<Smith.size()) && (Smith[lj] == Smith[ll])) {};
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
    } else {
        Inv.PushInvariant(1,ir);
    }
}

template<class InvClass>
void Smith_fly(InvClass& Inv, char * File_Name, const Integer& IN) {
    startActivity(GlobalCommentator, "N-Torsion", "Rank N-Torsion Computation", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    unsigned long rank;
    Integer Valence = (IN>0)?IN:-IN;
    // cerr << "File Name: " << File_Name << ", Valence: " << IN << endl;
    IntFactorDom<> IFD;

    SPBBlong::Domain_t LongDom;
    SPBBlong TheMat( LongDom );
    unsigned long ni, nj; TheMat.read(ni,nj,File_Name);
    // cerr << "Matrix: " << ni << "x" << nj << endl;

    std::vector<unsigned long> r;
    std::vector<unsigned long> e;
    std::vector<bool> t;
    unsigned long ir;
    unsigned long k, dimr = GIVMIN(ni,nj);
    unsigned long minr = dimr, maxr = 0;
    
    char * UT = new char[170]; 

    Integer p(2);

//     while ((iszero(IN % p)) || (! IFD.isprime(p)))
//         IFD.nextprimein(p);

    /*
    IFD.prevprime(p, 1+MAXPRIMESIZE - (ulong)lrand48()%(ulong)MAXPRIMESIZE/20);
    while ( IFD.iszero(IN % p) || (! IFD.isprime(p))  ) 
        IFD.prevprimein(p);
    */
    while ( IFD.iszero(IN % p) || (! IFD.isprime(p))  ) 
        IFD.nextprimein(p);


    // cerr << "maxr: " << maxr << ", p:" << p << " ( == " << (long)p << ")" << endl;
    one_gauss( maxr, File_Name, (GFqDomain::Residu_t)p);

    
//     GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT ) << "Integer rank (mod " << (GFqDomain::Residu_t)(p) << ") : " << maxr << endl;
    std::string s; Integer2string( s, p);
    sprintf(UT,"Integer rank (mod %s) : %ld", s.c_str(), maxr);
    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
    ir = maxr;
    
    unsigned long ssize = ir;
    std::vector<GFqDomain::Residu_t> Smith(ssize);
    for (std::vector<GFqDomain::Residu_t>::iterator is = Smith.begin(); is != Smith.end(); ++is)
        *is = 1;
   

    Integer pr=1;
    for( k=0 ; (Valence > 1) ; ++k) {
        
        Integer im; IFD.factor(im, Valence);
	while (! IFD.isprime(im) ) { Integer fim; IFD.factor(fim, im); im =fim; }
        long ie=0; for(; iszero( Valence % im ); ++ie ) {
            Valence = Valence / im;
        }        

        pr *= im;
        t.resize(k+1);
        r.resize(k+1);
        e.resize(k+1);
        if (im < Integer(MAXPRIMESIZE)) {
            t[k] = true;
            e[k] = ie;
            one_gauss( r[k], File_Name, (GFqDomain::Residu_t)(im));
            minr = GIVMIN(minr,r[k]);
            maxr = GIVMAX(maxr,r[k]);
//             GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT ) << "rank mod " << (GFqDomain::Residu_t)(im) << " : " << r[k] << endl;
            std::string s; Integer2string( s, im);
            sprintf(UT,"rank mod %s : %ld", s.c_str(), r[k]);
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        } else {
            t[k] = false;
            std::string s; Integer2string(s, im);
            sprintf(UT,"WARNING, %s is TOO BIG. Supposed full rank, WARNING.", s.c_str());
            activityReport(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_WARNING);
        }
   

        unsigned long l=0,l1=0, count_ie;
       if (t[k]) {
                count_ie = 0;
                l = r[k]; 
                GFqDomain::Residu_t p = (GFqDomain::Residu_t)(im);
                GFqDomain::Residu_t pp = p;
                while (l<ir) {
                    for(unsigned long lk=ir-l;lk;--lk)
                        Smith[ssize-lk] *= p;
                    pp *= p;
                    if (++count_ie < e[k]) {
                        one_pp_gauss(l, File_Name, pp,p);
                        maxr = GIVMAX(maxr, l);
                    } else
                        break;
//                     GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT ) << "pp rank mod " << pp << " : " << l << endl;
                    std::string s; Integer2string( s, Integer(pp) );
                    sprintf(UT,"pp rank mod %s : %ld", s.c_str(), l);
                    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
                }
        }
           
       InvClass TempInv( Inv.n_row(), Inv.n_col() );
       Smith2Invariant( ir, minr, TempInv, Smith );

       { 
           std::ostringstream ost("Current Smith Invariants = [ ");
           ost << TempInv << "]" << std::endl;
           std::string s = ost.str();
           activityReport(GlobalCommentator, s.c_str(), LVL_NORMAL, PARTIAL_RESULT);
       
//        GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT) << "Current Smith Invariants = [ " << TempInv << "]" << endl;
       }
       {
           std::ostringstream ost("Remaining to factor : ");
           ost << Valence << std::endl;
           std::string s = ost.str();
           activityReport(GlobalCommentator, s.c_str(), LVL_NORMAL, PARTIAL_RESULT);
           
//        GlobalCommentator.report( LVL_NORMAL, PARTIAL_RESULT) << "Remaining to factor : " << Valence << endl;
       }
       

    }

    Smith2Invariant( ir, minr, Inv, Smith );
        
    delete [] UT;
    stopActivity(GlobalCommentator, "Done.", LVL_BLABLA, INTERNAL_DESCRIPTION ); 
};




#endif 
